import enum
import os
import time
from contextlib import nullcontext
from datetime import datetime, timedelta
from typing import (TYPE_CHECKING, Any, BinaryIO, Collection, Literal,
                    Optional, TextIO)

from pydantic import TypeAdapter

from ..exceptions import MercutoClientException, MercutoHTTPException
from ..util import batched
from . import _PayloadType, _raise_for_response
from ._util import BaseModel, serialise_timedelta

if TYPE_CHECKING:
    from ..client import MercutoClient


class ChannelClassification(enum.Enum):
    PRIMARY = 'PRIMARY'
    PRIMARY_EVENT_AGGREGATE = 'PRIMARY_EVENT_AGGREGATE'
    EVENT_METRIC = 'EVENT_METRIC'
    SECONDARY = 'SECONDARY'


class Units(BaseModel):
    code: str
    name: str
    unit: Optional[str]


class Channel(BaseModel):
    code: str
    project: str
    units: Optional[Units]
    sampling_period: Optional[timedelta]
    classification: ChannelClassification
    label: str
    metric: Optional[str]
    source: Optional[str]
    aggregate: Optional[str]
    value_range_min: Optional[float]
    value_range_max: Optional[float]
    multiplier: float
    offset: float
    last_valid_timestamp: Optional[datetime]
    is_wallclock_interval: bool


class Expression(BaseModel):
    expression: str
    target: Channel


class DatatableColumn(BaseModel):
    channel: str
    column_label: str


class Datatable(BaseModel):
    code: str
    project: str
    name: str
    enabled: bool
    sampling_period: Optional[timedelta] = None
    columns: list[DatatableColumn]


class SecondaryDataSample(BaseModel):
    channel: str
    timestamp: datetime
    value: float


class MetricDataSample(BaseModel):
    channel: str
    timestamp: datetime
    value: float
    event: str


class LatestDataSample(BaseModel):
    channel: str
    timestamp: datetime
    value: float


_ChannellistAdapter = TypeAdapter(list[Channel])
_ExpressionlistAdapter = TypeAdapter(list[Expression])
_DatatablelistAdapter = TypeAdapter(list[Datatable])
_UnitslistAdapter = TypeAdapter(list[Units])
_MetricSamplelistAdapter = TypeAdapter(list[MetricDataSample])
_SecondarySamplelistAdapter = TypeAdapter(list[SecondaryDataSample])
_LatestSampleListAdapter = TypeAdapter(list[LatestDataSample])


class FrameFormat(enum.Enum):
    COLUMNS = "COLUMNS"
    SAMPLES = "SAMPLES"


class FileFormat(enum.Enum):
    FEATHER = "FEATHER"
    PARQUET = "PARQUET"
    CSV = "CSV"


class ChannelFormat(enum.Enum):
    CODE = "CODE"
    LABEL = "LABEL"


class AggregationOptions(BaseModel):
    method: Literal['min', 'max', 'mean', 'sum', 'count', 'greatest']
    interval: Literal['second', 'minute', 'hour', 'day', 'week', 'month', 'year']
    rolling: bool = False


class GetStatusRequestResponse(BaseModel):
    class GetDataRequestStatusCompletedResult(BaseModel):
        class ResultMetadata(BaseModel):
            first_timestamp: Optional[datetime]
        result_url: str
        expires_at: datetime
        mime_type: str
        file_size: int
        metadata: ResultMetadata

    request_id: str
    status_code: int
    message: str
    requested_at: Optional[datetime]
    completed_at: Optional[datetime]
    result: Optional["GetStatusRequestResponse.GetDataRequestStatusCompletedResult"]


class Healthcheck(BaseModel):
    status: str


class MercutoDataService:
    def __init__(self, client: 'MercutoClient', path: str = '/v2/data') -> None:
        self._client = client
        self._path = path

    def healthcheck(self) -> Healthcheck:
        r = self._client._http_request(f"{self._path}/healthcheck", "GET")
        return Healthcheck.model_validate_json(r.text)

    def refresh_continuous_aggregates(self) -> None:
        """
        Request a refresh of continuous aggregates on all tables
        """
        self._client._http_request(f"{self._path}/meta/refresh-aggregates", "POST")

    """
    Channels
    """

    def list_channels(self, project: str, classification: Optional[ChannelClassification] = None,
                      aggregate: Optional[str] = None, metric: Optional[str] = None,
                      show_hidden: bool = False) -> list[Channel]:
        params: dict[str, Any] = {
            'project': project,
            'limit': 100,
            'offset': 0,
            'show_hidden': show_hidden,
        }
        if classification:
            params['classification'] = classification.value
        if aggregate:
            params['aggregate'] = aggregate
        if metric:
            params['metric'] = metric

        all_channels = []
        while True:
            r = self._client._http_request(f'{self._path}/channels', 'GET', params=params)

            channels = _ChannellistAdapter.validate_json(r.text)
            all_channels.extend(channels)
            if len(channels) < params['limit']:
                break
            params['offset'] += params['limit']
        return all_channels

    def get_channel(self, code: str) -> Optional[Channel]:
        r = self._client._http_request(f'{self._path}/channels/{code}', 'GET', raise_for_status=False)
        if r.status_code == 404:
            return None
        _raise_for_response(r)
        return Channel.model_validate_json(r.text)

    def update_channel(self, code: str, label: Optional[str] = None, units: Optional[str] = None,
                       metric: Optional[str] = None, multiplier: Optional[float] = None,
                       offset: Optional[float] = None) -> Channel:
        payload: _PayloadType = {}
        if label is not None:
            payload['label'] = label
        if units is not None:
            payload['units'] = units
        if metric is not None:
            payload['metric'] = metric
        if multiplier is not None:
            payload['multiplier'] = multiplier
        if offset is not None:
            payload['offset'] = offset

        r = self._client._http_request(f'{self._path}/channels/{code}', 'PATCH', json=payload)
        return Channel.model_validate_json(r.text)

    def delete_channel(self, code: str) -> bool:
        r = self._client._http_request(f'{self._path}/channels/{code}', 'DELETE')
        return r.status_code == 204

    def create_channel(self, project: str,
                       label: str,
                       classification: ChannelClassification = ChannelClassification.SECONDARY,
                       sampling_period: Optional[timedelta] = None,
                       multiplier: float = 1.0, offset: float = 0.0,
                       value_range_min: Optional[float] = None, value_range_max: Optional[float] = None,
                       delta_max: Optional[float] = None,
                       units: Optional[str] = None,
                       aggregate: Optional[str] = None,
                       source: Optional[str] = None,
                       metric: Optional[str] = None) -> Channel:
        payload: _PayloadType = {
            'project': project,
            'label': label,
            'classification': classification.value,
            'multiplier': multiplier,
            'offset': offset,
        }
        if sampling_period is not None:
            payload['sampling_period'] = serialise_timedelta(sampling_period)
        if value_range_min is not None:
            payload['value_range_min'] = value_range_min
        if value_range_max is not None:
            payload['value_range_max'] = value_range_max
        if delta_max is not None:
            payload['delta_max'] = delta_max
        if units is not None:
            payload['units'] = units
        if aggregate is not None:
            payload['aggregate'] = aggregate
        if source is not None:
            payload['source'] = source
        if metric is not None:
            payload['metric'] = metric

        r = self._client._http_request(f'{self._path}/channels', 'PUT', json=payload)
        return Channel.model_validate_json(r.text)

    """
    Expressions
    """

    def create_expression(
        self,
        project: str,
        label: str,
        expression: str,
        units: Optional[str] = None,
        aggregate: Optional[str] = None,
        metric: Optional[str] = None
    ) -> Expression:
        payload: _PayloadType = {
            "project": project,
            "label": label,
            "expression": expression,
        }
        if units is not None:
            payload["units"] = units
        if aggregate is not None:
            payload["aggregate"] = aggregate
        if metric is not None:
            payload["metric"] = metric

        r = self._client._http_request(f'{self._path}/expressions', 'PUT', json=payload)
        return Expression.model_validate_json(r.text)

    def delete_expression(self, code: str) -> bool:
        r = self._client._http_request(f'{self._path}/expressions/{code}', 'DELETE')
        return r.status_code == 202

    """
    Datatables
    """

    def create_datatable(self, project: str, name: str, sampling_period: timedelta, column_labels: Collection[str]) -> Datatable:
        payload: _PayloadType = {
            "project": project,
            "name": name,
            "sampling_period": serialise_timedelta(sampling_period),
            "column_labels": list(column_labels),
        }
        r = self._client._http_request(f'{self._path}/datatables', 'PUT', json=payload)
        return Datatable.model_validate_json(r.text)

    def list_datatables(self, project: str) -> list[Datatable]:
        datatables = []
        params: dict[str, Any] = {
            "project": project,
            "limit": 100,
            "offset": 0,
        }
        while True:
            r = self._client._http_request(f'{self._path}/datatables', 'GET', params=params)

            batch = _DatatablelistAdapter.validate_json(r.text)
            datatables.extend(batch)
            if len(batch) < params["limit"]:
                break
            params["offset"] += params["limit"]
        return datatables

    """
    Units
    """

    def list_units(self) -> list[Units]:
        r = self._client._http_request(f'{self._path}/units', 'GET')
        return _UnitslistAdapter.validate_json(r.text)

    def create_unit(self, name: str, unit: str) -> Units:
        payload: _PayloadType = {
            "name": name,
            "unit": unit,
        }
        r = self._client._http_request(f'{self._path}/units', 'PUT', json=payload)
        return Units.model_validate_json(r.text)

    """
    Requests
    """

    def create_request(
        self,
        start_time: datetime,
        end_time: datetime,
        project: Optional[str] = None,
        channels: Optional[Collection[str]] = None,
        classification: Optional[ChannelClassification] = None,
        frame_format: FrameFormat = FrameFormat.SAMPLES,
        file_format: FileFormat = FileFormat.PARQUET,
        channel_format: ChannelFormat = ChannelFormat.CODE,
        aggregation: Optional[AggregationOptions] = None,
        timeout: float = 0
    ) -> GetStatusRequestResponse:
        if timeout > 20:
            timeout = 20  # Cap timeout to 20 seconds

        if channels is None and classification is None:
            raise ValueError("Must supply either channels or classification.")

        payload: _PayloadType = {
            "start_time": start_time.isoformat(),
            "end_time": end_time.isoformat(),
            "frame_format": frame_format.value,
            "file_format": file_format.value,
            "channel_format": channel_format.value,
        }

        if project:
            payload["project"] = project

        if channels:
            payload["channels"] = list(channels)

        if classification:
            payload["classification"] = classification.value

        if aggregation is not None:
            payload["aggregation"] = aggregation.model_dump(mode='json')

        r = self._client._http_request(
            f'{self._path}/requests', 'POST',
            json=payload,
            params={"timeout": timeout}
        )
        return GetStatusRequestResponse.model_validate_json(r.text)

    def get_request(self, request_id: str) -> GetStatusRequestResponse:
        r = self._client._http_request(f'{self._path}/requests/{request_id}', 'GET')
        return GetStatusRequestResponse.model_validate_json(r.text)

    """
    Request Helpers
    """

    def load_presigned_url(
        self,
        start_time: datetime,
        end_time: datetime,
        project: Optional[str] = None,
        channels: Optional[Collection[str]] = None,
        classification: Optional[ChannelClassification] = None,
        frame_format: FrameFormat = FrameFormat.SAMPLES,
        file_format: FileFormat = FileFormat.PARQUET,
        channel_format: ChannelFormat = ChannelFormat.CODE,
        aggregation: Optional[AggregationOptions] = None,
        poll_interval: float = 0.25,
        timeout: int = 60
    ) -> str:
        """
        Request a presigned download URL for data and poll until ready.

        Returns:
            The presigned result_url as a string.
        Raises:
            MercutoHTTPException, MercutoClientException on error or timeout.
        """

        result = self.load_data_request(
            start_time=start_time,
            end_time=end_time,
            project=project,
            channels=channels,
            classification=classification,
            frame_format=frame_format,
            file_format=file_format,
            channel_format=channel_format,
            aggregation=aggregation,
            poll_interval=poll_interval,
            timeout=timeout
        )
        if result.result_url is None:
            raise MercutoClientException("Failed to obtain presigned URL.")
        return result.result_url

    def load_data_request(
        self,
        start_time: datetime,
        end_time: datetime,
        project: Optional[str] = None,
        channels: Optional[Collection[str]] = None,
        classification: Optional[ChannelClassification] = None,
        frame_format: FrameFormat = FrameFormat.SAMPLES,
        file_format: FileFormat = FileFormat.PARQUET,
        channel_format: ChannelFormat = ChannelFormat.CODE,
        aggregation: Optional[AggregationOptions] = None,
        poll_interval: float = 0.25,
        timeout: int = 60
    ) -> GetStatusRequestResponse.GetDataRequestStatusCompletedResult:
        """
        Request a presigned download URL for data and poll until ready.

        Returns:
            The GetStatusRequestResponse
        Raises:
            MercutoHTTPException, MercutoClientException on error or timeout.
        """

        # Start the request, using poll_interval as the initial timeout
        status = self.create_request(
            project=project,
            channels=channels,
            start_time=start_time,
            end_time=end_time,
            classification=classification,
            frame_format=frame_format,
            file_format=file_format,
            channel_format=channel_format,
            timeout=poll_interval,
            aggregation=aggregation
        )
        request_id = status.request_id

        # If already complete, return immediately
        if status.status_code == 200 and status.result and status.result.result_url:
            return status.result
        if status.status_code >= 400:
            raise MercutoHTTPException(status.message, status.status_code)

        # Otherwise, poll for completion
        start_poll = time.time()
        while True:
            status = self.get_request(request_id)
            if status.status_code == 200 and status.result and status.result.result_url:
                return status.result
            if status.status_code >= 400:
                raise MercutoHTTPException(status.message, status.status_code)
            if time.time() - start_poll > timeout:
                raise MercutoClientException("Timed out waiting for presigned url.")
            time.sleep(poll_interval)

    """
    Samples
    """

    def insert_secondary_samples(
        self,
        project: str,
        samples: Collection[SecondaryDataSample]
    ) -> None:
        """
        Insert secondary samples.
        """
        for batch in batched(samples, 5000):
            payload = _SecondarySamplelistAdapter.dump_python(list(batch), mode='json')
            self._client._http_request(
                f'{self._path}/samples/secondary', 'PUT', json=payload, params={"project": project}
            )

            # No return value, 202 accepted

    def insert_metric_samples(
        self,
        project: str,
        samples: Collection[MetricDataSample]
    ) -> None:
        """
        Insert metric samples.
        """
        for batch in batched(samples, 5000):
            payload = _MetricSamplelistAdapter.dump_python(list(batch), mode='json')
            self._client._http_request(
                f'{self._path}/samples/metric', 'PUT', json=payload, params={"project": project}
            )

        # No return value, 202 accepted

    def load_secondary_samples(
        self,
        channels: Collection[str],
        start_time: datetime,
        end_time: datetime,
        limit: int = 100
    ) -> list[SecondaryDataSample]:
        """
        Load up to 100 secondary samples.
        """
        params: _PayloadType = {
            "channels": list(channels),
            "start_time": start_time.isoformat(),
            "end_time": end_time.isoformat(),
            "limit": limit
        }
        r = self._client._http_request(
            f'{self._path}/samples/secondary', 'GET', params=params
        )

        return _SecondarySamplelistAdapter.validate_json(r.text)

    def load_metric_samples(
        self,
        channels: Optional[Collection[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        events: Optional[Collection[str]] = None,
        project: Optional[str] = None,
        limit: int = 100
    ) -> list[MetricDataSample]:
        """
        Load up to 100 metric samples.
        """
        params: _PayloadType = {
            "limit": limit
        }
        if project is not None:
            params["project"] = project
        if channels is not None:
            params["channels"] = list(channels)
        if start_time is not None:
            params["start_time"] = start_time.isoformat()
        if end_time is not None:
            params["end_time"] = end_time.isoformat()
        if events is not None:
            params["event"] = list(events)
        r = self._client._http_request(
            f'{self._path}/samples/metric', 'GET', params=params
        )

        return _MetricSamplelistAdapter.validate_json(r.text)

    def load_metric_sample(self, channel: str, event: str) -> Optional[float]:
        """
        Load a single metric sample for a specific channel and event.
        """
        samples = self.load_metric_samples([channel], events=[event])
        return samples[0].value if samples else None

    def delete_metric_samples(self, project: str, event: str, channels: Optional[Collection[str]] = None) -> None:
        params: _PayloadType = {"project": project, "event": event}
        if channels is not None:
            params["channels"] = list(channels)
        self._client._http_request(
            f'{self._path}/samples/metric', 'DELETE', params=params
        )

    def upload_file(self, project: str, datatable: str, file: str | bytes | TextIO | BinaryIO,
                    filename: Optional[str] = None,
                    timezone: Optional[str] = None) -> None:
        if isinstance(file, str):
            ctx = open(file, 'rb')
            filename = filename or os.path.basename(file)
        else:
            ctx = nullcontext(file)  # type: ignore
            filename = filename or 'file.dat'

        params: _PayloadType = {
            "project": project,
            "datatable": datatable,
        }
        if timezone is not None:
            params["timezone"] = timezone

        with ctx as f:
            self._client._http_request(f'{self._path}/files/upload/small', 'POST',
                                       params=params,
                                       files={'file': (filename, f, 'text/csv')})

    def get_latest_samples(self, project: str, include_primary: bool = True) -> list[LatestDataSample]:
        params: _PayloadType = {
            "project": project,
            "include_primary": include_primary
        }
        r = self._client._http_request(
            f'{self._path}/statistics/latest-samples', 'GET', params=params
        )

        return _LatestSampleListAdapter.validate_json(r.text)



import logging
import os
import shutil
import sqlite3
from datetime import datetime, timezone
from typing import Callable, Optional

logger = logging.getLogger(__name__)


def _default_standard_clock(_: str):
    # Default clock function to get current time in seconds since epoch
    return datetime.now(timezone.utc).timestamp()


def _default_file_clock(filepath: str) -> float:
    # Default clock function to get file creation time in seconds since epoch
    return os.path.getctime(filepath)


class FileProcessor:
    """
    System for processing files in a strict order with retry logic.

    Keeps an SQLite database to track files and their processing status.
    Keeps old files in the buffer and only deletes them once max_files is reached (whether processed or not).

    :param buffer_dir: Directory where files are stored
    :param db_path: Path to the SQLite database file for tracking file processing status.
    :param process_callback: Callable that processes a file. Should return True if processing is successful
    :param max_attempts: Maximum number of attempts to process a file before marking it as failed.
    :param max_files: Maximum number of files to keep in the buffer directory. If None, no limit is enforced.
    :param free_space_mb: Optional minimum free space in MB to keep on the partition where the buffer directory is located.
        This is combined with max_files to determine when to delete old files and takes precedence over max_files.
    :param clock: Optional callable that returns the timestamp for the file based on the filename. Takes in the file name as an argument
        and should return a float representing the timestamp in seconds since the epoch. Defaults to datetime.now().timestamp().
        This clock is NOT used when scanning existing files, it takes its own clock function.


    Provides a callback for processing files, which should return True if successful.
    If processing fails, it retries up to max_attempts times before marking the file as failed.

    Periodically call `cleanup_old_files()` to remove old files from the buffer directory.
    Use `scan_existing_files()` to register files that were added while the system was offline.
    Use `process_next_file()` to process the next file in the buffer in strict order.
    Add files to the buffer using `add_file_to_db()`, or use `start_watching()` to automatically watch a directory for new files.

    Example usage:
    ```python
    processor = FileProcessor(...)
    processor.scan_existing_files()

    while True:
        processor.add_file_to_db("path/to/file.txt")
        processor.process_next_file()
    ```
    """

    def __init__(self, buffer_dir: str, db_path: str,
                 process_callback: Callable[[str], bool],
                 max_attempts: int,
                 max_files: Optional[int] = None,
                 free_space_mb: Optional[float] = None,
                 clock: Optional[Callable[[str], float]] = None
                 ) -> None:
        self._buffer_dir = buffer_dir
        self._db_path = db_path
        self._max_files = max_files
        self._max_attempts = max_attempts
        self._process_callback = process_callback
        self._free_space_mb = free_space_mb
        self._clock = clock if clock is not None else _default_standard_clock
        os.makedirs(self._buffer_dir, exist_ok=True)
        self._init_db()

    def get_db_path(self) -> str:
        """Returns the path to the SQLite database."""
        return self._db_path

    def get_buffer_dir(self) -> str:
        """Returns the path to the buffer directory."""
        return self._buffer_dir

    def _init_db(self) -> None:
        """Initialize SQLite database with attempt tracking."""
        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()
        cursor.execute("""
        CREATE TABLE IF NOT EXISTS file_buffer (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            filename TEXT UNIQUE,
            filepath TEXT UNIQUE,
            status TEXT DEFAULT 'pending',
            attempts INTEGER DEFAULT 0,
            timestamp REAL
        )
        """)
        conn.commit()
        conn.close()
        logger.info("Database initialized.")

    def scan_existing_files(self, clock: Optional[Callable[[str], float]] = None) -> None:
        """
        Detect files added while offline and process them.
        :param clock: Optional callable that returns the timestamp for the file based on the filename.
            If not provided, uses the file's creation time.
            This may be innaccurate if multiple files are added at once, or if the file system does not support accurate timestamps.

        """
        if clock is None:
            clock = _default_file_clock

        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()

        files_with_timestamps = [
            (filename, clock(os.path.join(self._buffer_dir, filename)))
            for filename in os.listdir(self._buffer_dir)
            if os.path.isfile(os.path.join(self._buffer_dir, filename))
        ]

        for filename, timestamp in sorted(files_with_timestamps, key=lambda x: x[1]):
            filepath = os.path.join(self._buffer_dir, filename)
            timestamp = clock(filepath)

            cursor.execute(
                "SELECT COUNT(*) FROM file_buffer WHERE filename = ?", (filename,))
            exists: int = cursor.fetchone()[0]

            if not exists:
                logger.info(f"Registering existing {filename} for processing...")
                cursor.execute("INSERT INTO file_buffer (filename, filepath, status, attempts, timestamp) VALUES (?, ?, 'pending', 0, ?)",
                               (filename, filepath, timestamp))

        conn.commit()
        conn.close()

    def process_next_file(self) -> Optional[str]:
        """
        Attempt to process the next file in the sequence (if exists), ensuring strict order.
        Returns the filepath of the processed file if successful or None if no pending files are found or failed.
        """
        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()
        cursor.execute(
            "SELECT filepath, attempts FROM file_buffer WHERE status = 'pending' ORDER BY timestamp ASC LIMIT 1")

        pending_files: list[tuple[str, int]] = cursor.fetchall()
        conn.close()

        assert len(pending_files) <= 1, "More than one pending file found, which violates strict order."

        for (filepath, attempts) in pending_files:
            if self._process_file(filepath, attempts):
                return filepath
        return None

    def _process_file(self, filepath: str, attempts: int) -> bool:
        if not os.path.exists(filepath):
            logger.warning(f"File {filepath} does not exist. Skipping.")
            self._mark_as_failed(filepath)
        try:
            success: bool = self._process_callback(filepath)
        except Exception as e:
            logger.error(f"Processing error for {filepath}: {e}")
            success = False

        if success:
            self._mark_as_processed(filepath)
            return True

        if attempts >= self._max_attempts:
            logger.warning(
                f"Max retries reached for {filepath}. Moving to next file.")

            self._mark_as_failed(filepath)
            return True  # Give up and move to next file
        else:
            self._increment_attempt(filepath)
            return False

    def _increment_attempt(self, filepath: str) -> None:
        # Reopen DB to update attempt count
        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()
        cursor.execute(
            "UPDATE file_buffer SET attempts = attempts + 1 WHERE filepath = ?", (filepath,))
        conn.commit()
        conn.close()

    def _mark_as_failed(self, filepath: str) -> None:
        """Marks a file as failed in the database."""
        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()
        cursor.execute(
            "UPDATE file_buffer SET status = 'failed' WHERE filepath = ?", (filepath,))
        conn.commit()
        conn.close()
        logger.info(f"File {filepath} marked as failed.")

    def _mark_as_processed(self, filepath: str) -> None:
        """Marks a file as processed in the database."""
        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()
        cursor.execute(
            "UPDATE file_buffer SET status = 'processed' WHERE filepath = ?", (filepath,))
        conn.commit()
        conn.close()
        logger.info(f"File {filepath} marked as processed.")

    def cleanup_old_files_with_max_files(self) -> None:
        """Remove old files beyond the max file count."""
        if self._max_files is None:
            return
        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()
        cursor.execute(
            "SELECT filepath FROM file_buffer ORDER BY timestamp DESC LIMIT -1 OFFSET ?", (self._max_files,))
        files_to_delete: list[tuple[str]] = cursor.fetchall()
        conn.close()

        for (filepath,) in reversed(files_to_delete):
            if os.path.exists(filepath):
                os.remove(filepath)
                conn = sqlite3.connect(self._db_path)
                cursor = conn.cursor()
                cursor.execute(
                    "DELETE FROM file_buffer WHERE filepath = ?", (filepath,))
                conn.commit()
                conn.close()
                logger.info(f"Deleted old file {filepath}")

    def cleanup_old_files(self) -> None:
        """Remove old files based on max_files and free space."""
        if self._max_files is not None:
            self.cleanup_old_files_with_max_files()

        if self._free_space_mb is not None:
            self.cleanup_old_files_with_free_space()

    def cleanup_old_files_with_free_space(self) -> None:
        """Remove old files ensuring free space is maintained."""
        if self._free_space_mb is None:
            return

        def free_space_mb() -> float:
            """Returns the free space in MB on the partition where the buffer directory is located."""
            _, _, free = shutil.disk_usage(self._buffer_dir)
            return free / (1024 * 1024)

        while free_space_mb() < self._free_space_mb:
            if not self._delete_oldest_file():
                logger.warning("No more files to delete to free up space.")
                break

    def _delete_oldest_file(self) -> bool:
        """Deletes the oldest file in the buffer directory."""
        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()
        cursor.execute(
            "SELECT filepath FROM file_buffer ORDER BY timestamp ASC LIMIT 1")
        oldest_file: Optional[tuple[str]] = cursor.fetchone()
        if not oldest_file:
            logger.info("No files to delete.")
            return False

        filepath = oldest_file[0]
        if os.path.exists(filepath):
            os.remove(filepath)
            cursor.execute(
                "DELETE FROM file_buffer WHERE filepath = ?", (filepath,))
            conn.commit()
            logger.info(f"Deleted oldest file {filepath}")
        else:
            logger.warning(f"Oldest file {filepath} does not exist.")

        conn.close()
        return True

    def add_file_to_db(self, filepath: str) -> None:
        """Adds a new file to database and triggers processing."""
        timestamp = self._clock(filepath)
        filename: str = os.path.basename(filepath)
        conn = sqlite3.connect(self._db_path)
        cursor = conn.cursor()

        cursor.execute("""
        INSERT OR IGNORE INTO file_buffer (filename, filepath, status, attempts, timestamp)
        VALUES (?, ?, 'pending', 0, ?)
        """, (filename, filepath, timestamp))

        conn.commit()
        conn.close()



import ftplib
import io
import os
import tempfile
from datetime import datetime, timezone

from ...ingester.ftp import simple_ftp_server


def test_simple_ftp_server():
    receives = []
    def clock(): return datetime(2023, 10, 1, 12, 0, 0, tzinfo=timezone.utc)

    with tempfile.TemporaryDirectory() as temp_dir:
        with simple_ftp_server(directory=temp_dir,
                               username='test', password='password', port=2121,
                               callback=lambda dest: receives.append(dest), clock=clock):
            client = ftplib.FTP()
            client.connect('localhost', 2121)
            client.login('test', 'password')

            # Upload a test file
            testbuf = io.BytesIO(b'This is a test file.')
            client.storbinary('STOR test_file.txt', testbuf)

        assert len(receives) == 1
        assert receives[0].endswith('test_file_20231001T120000.txt')

        # Verify the file was uploaded correctly
        with open(receives[0], 'rb') as f:
            content = f.read()
            assert content == b'This is a test file.'

        # Ensure that the file exists in the temp directory
        found = os.listdir(temp_dir)
        assert 'test_file_20231001T120000.txt' in found
        assert len(found) == 1  # Only one file should be present



from datetime import timedelta

from pydantic import BaseModel as _BaseModel
from pydantic import ConfigDict, TypeAdapter

_TimedeltaAdapter = TypeAdapter(timedelta)


def serialise_timedelta(td: timedelta) -> str:
    s = _TimedeltaAdapter.dump_python(td, mode='json')
    assert isinstance(s, str)
    return s


class BaseModel(_BaseModel):
    model_config = ConfigDict(
        extra='allow'
    )



from typing import Optional

import pytz

from ...modules.data import SecondaryDataSample
from .generic_csv import parse_generic_csv_file


def parse_worldsensing_standard_file(filename: str, label_to_channel_code: dict[str, str],
                                     timezone: Optional[pytz.BaseTzInfo] = None) -> list[SecondaryDataSample]:
    """
    Parse a worldsensing standard CSV file provided when downloading data or using standard CSV export.
    """
    return parse_generic_csv_file(
        filename, label_to_channel_code, header_index=9, data_index=0, timezone=timezone)


def parse_worldsensing_compact_file(filename: str, label_to_channel_code: dict[str, str],
                                    timezone: Optional[pytz.BaseTzInfo] = None) -> list[SecondaryDataSample]:
    """
    Parse a worldsensing custom CSV file. These are generated when using compacted CSV mechanism.
    """
    return parse_generic_csv_file(
        filename, label_to_channel_code, header_index=1, data_index=0, timezone=timezone)



import argparse
import logging
import logging.handlers
import os
import sys
import time
from typing import Callable, TypeVar

import schedule

from ..util import get_free_space_excluding_files
from .ftp import simple_ftp_server
from .mercuto import MercutoIngester
from .processor import FileProcessor

logger = logging.getLogger(__name__)


T = TypeVar('T')


def call_and_log_error(func: Callable[[], T]) -> T | None:
    """
    Call a function and log any exceptions that occur.
    """
    try:
        return func()
    except Exception:
        logging.exception(f"Error in {func.__name__}")
        return None


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Mercuto Ingester CLI')
    parser.add_argument('-p', '--project', type=str,
                        required=True, help='Mercuto project code')
    parser.add_argument('-k', '--api-key', type=str,
                        required=True, help='API key for Mercuto')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Enable verbose output')
    parser.add_argument('-d', '--directory', type=str,
                        help='Directory to store ingested files. Default is a directory called `buffered-files` in the workdir.')
    parser.add_argument('-s', '--size', type=int,
                        help='Size in MB for total amount of files to store in the buffer. \
                            Default is 75% of the available disk space on the buffer partition excluding the directory itself', default=None)
    parser.add_argument('--max-files', type=int,
                        help='Maximum number of files to keep in the buffer. Default is to use the size param.', default=None)
    parser.add_argument('--max-attempts', type=int,
                        help='Maximum number of attempts to process a file before giving up. Default is 1000.',
                        default=1000)
    parser.add_argument('--workdir', type=str,
                        help='Working directory for the ingester. Default is ~/.mercuto-ingester',)
    parser.add_argument('--logfile', type=str,
                        help='Log file path. No logs written if not provided. Maximum of 4 log files of 1MB each will be kept.\
                            Default is log.txt in the workdir.')
    parser.add_argument('--mapping', type=str,
                        help='Path to a JSON file with channel label to channel code mapping.\
                            If not provided, the ingester will try to detect the channels from the project.',
                        default=None)
    parser.add_argument('--hostname', type=str,
                        help='Hostname to use for the Mercuto server. Default is "https://api.rockfieldcloud.com.au".',
                        default='https://api.rockfieldcloud.com.au')
    parser.add_argument('--clean',
                        help='Drop the database before starting. This will not remove any buffer files and will rescan them on startup.',
                        action='store_true')
    parser.add_argument('--username', type=str,
                        help='Username for the FTP server. Default is "logger".',
                        default='logger')
    parser.add_argument('--password', type=str,
                        help='Password for the FTP server. Default is "password".',
                        default='password')
    parser.add_argument('--port', type=int,
                        help='Port for the FTP server. Default is 2121.',
                        default=2121)
    parser.add_argument('--no-rename', action='store_true',
                        help='Add the current timestamp to the end of the files received via FTP. \
                        This is useful to avoid overwriting files with the same name.')

    args = parser.parse_args()

    if args.workdir is None:
        workdir = os.path.join(os.path.expanduser('~'), ".mercuto-ingester")
    else:
        workdir = args.workdir
        if not os.path.exists(args.workdir):
            raise ValueError(f"Work directory {args.workdir} does not exist")
    os.makedirs(workdir, exist_ok=True)

    if args.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO

    handlers: list[logging.Handler] = []
    handlers.append(logging.StreamHandler(sys.stderr))

    if args.logfile is not None:
        logfile = args.logfile
    else:
        logfile = os.path.join(workdir, 'log.txt')
    handlers.append(logging.handlers.RotatingFileHandler(
        logfile, maxBytes=1000000, backupCount=3))

    logging.basicConfig(format='[PID %(process)d] %(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                        datefmt='%d/%m/%Y %H:%M:%S',
                        level=level,
                        handlers=handlers)

    if args.directory is None:
        buffer_directory = os.path.join(workdir, "buffered-files")
    else:
        buffer_directory = args.directory
    os.makedirs(buffer_directory, exist_ok=True)

    ftp_dir = os.path.join(workdir, 'temp-ftp-data')
    os.makedirs(ftp_dir, exist_ok=True)

    size = args.size
    if size is None and args.max_files is None:
        size = get_free_space_excluding_files(buffer_directory) * 0.75 // (1024 * 1024)  # Convert to MB
        logging.info(f"Buffer size set to {size} MB based on available disk space.")

    if args.mapping is not None:
        import json
        with open(args.mapping, 'r') as f:
            mapping = json.load(f)
        if not isinstance(mapping, dict):
            raise ValueError(f"Mapping file {args.mapping} must contain a JSON object")
    else:
        mapping = {}

    logger.info(f"Using work directory: {workdir}")

    database_path = os.path.join(workdir, "buffer.db")
    if args.clean and os.path.exists(database_path):
        logging.info(f"Dropping existing database at {database_path}")
        os.remove(database_path)

    ingester = MercutoIngester(
        project_code=args.project,
        api_key=args.api_key,
        hostname=args.hostname)

    ingester.update_mapping(mapping)

    processor = FileProcessor(
        buffer_dir=buffer_directory,
        db_path=database_path,
        process_callback=ingester.process_file,
        max_attempts=args.max_attempts,
        free_space_mb=size,
        max_files=args.max_files)

    processor.scan_existing_files()

    with simple_ftp_server(directory=buffer_directory,
                           username=args.username, password=args.password, port=args.port,
                           callback=processor.add_file_to_db, rename=not args.no_rename,
                           workdir=ftp_dir):
        schedule.every(60).seconds.do(call_and_log_error, ingester.ping)
        schedule.every(5).seconds.do(call_and_log_error, processor.process_next_file)
        schedule.every(2).minutes.do(call_and_log_error, processor.cleanup_old_files)

        while True:
            schedule.run_pending()
            time.sleep(0.5)



import contextlib
import logging
import os
import shutil
import tempfile
import threading
from datetime import datetime, timezone
from typing import Callable, Iterator, Optional

from pyftpdlib import authorizers  # type: ignore[import-untyped]
from pyftpdlib.handlers import FTPHandler  # type: ignore[import-untyped]
from pyftpdlib.servers import FTPServer  # type: ignore[import-untyped]

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def simple_ftp_server(directory: str,
                      username: str,
                      password: str,
                      port: int = 2121,
                      callback: Optional[Callable[[str], None]] = None,
                      workdir: Optional[str] = None,
                      rename: bool = True,
                      clock: Optional[Callable[[], datetime]] = None) -> Iterator[None]:
    """
    Wrapper for a simple FTP server that allows uploading files to a specified directory.
    Callback function can be provided which is called with the destination path of each uploaded file.
    Files are first uploaded to a workdirectory and then moved to the specified directory.
    If workdir is not specified, a temporary directory is used.

    :param directory: Directory where files will be uploaded.
    :param username: Username for FTP authentication.
    :param password: Password for FTP authentication.
    :param port: Port on which the FTP server will listen.
    :param callback: Optional callback function that is called with the destination path of each uploaded file.
    :param workdir: Optional working directory where files are initially uploaded before moving to the final directory.
    :param rename: If True, appends a timestamp to the filename to avoid overwriting existing files.
    :param clock: Function to get the current time, defaults to datetime.now with timezone UTC
    :return: Context manager that starts the FTP server and allows file uploads.

    Runs in a background thread during context manager usage.

    Example usage:

    ```python
    def my_callback(dest_path: str):
        print(f"File uploaded to: {dest_path}")

    with simple_ftp_server('/path/to/upload/dir', 'user', 'pass', port=2121, callback=my_callback) as server:
        # Your code here, e.g. processing files
        while True:
            time.sleep(10)
    ```
    """

    if clock is None:
        def clock(): return datetime.now(timezone.utc)

    def rename_file(file_path: str) -> str:
        """
        Rename the file by appending a timestamp to avoid overwriting.
        Adds the timestamp before the file extension.
        """
        base, ext = os.path.splitext(file_path)
        timestamp = clock().strftime("%Y%m%dT%H%M%S")
        new_name = f"{base}_{timestamp}{ext}"
        return new_name

    class CustomFTPHandler(FTPHandler):
        def on_file_received(self, file):
            target = os.path.join(directory, os.path.basename(file))

            if rename:
                target = rename_file(target)

            dest = shutil.move(file, target)
            if callback:
                callback(dest)

        def on_incomplete_file_received(self, file):
            try:
                os.remove(file)
            except Exception:
                logger.exception(f"Failed to remove incomplete file: {file}")

    if workdir is None:
        workdir_ctx: contextlib.AbstractContextManager[str] = tempfile.TemporaryDirectory(prefix="ftp_workdir_")
    else:
        workdir_ctx = contextlib.nullcontext(workdir)

    with workdir_ctx as workdir:
        authorizer = authorizers.DummyAuthorizer()
        authorizer.add_user(username, password,
                            workdir, perm='lwe')
        handler = CustomFTPHandler
        handler.authorizer = authorizer
        handler.banner = "FTP Server Ready."
        handler.passive_ports = range(60000, 65535)

        address = ('0.0.0.0', port)
        server = FTPServer(address, handler)
        server.max_cons = 60
        server.max_cons_per_ip = 20

        server_thread = threading.Thread(target=server.serve_forever, daemon=True)
        logger.debug(f"Starting FTP server on {port}...")
        server_thread.start()
        try:
            yield
        finally:
            logger.debug("Stopping FTP server...")
            server.close_all()
            server_thread.join(timeout=10)
            logger.debug("FTP server stopped.")



import contextlib
import json as json_stdlib
import logging
import os
import time
from typing import (Any, Iterator, Literal, Mapping, Optional, Protocol, Type,
                    TypeVar)

import requests
import requests.cookies

from ._authentication import (IAuthenticationMethod,
                              create_authentication_method)
from .exceptions import MercutoClientException, MercutoHTTPException
from .modules.core import MercutoCoreService
from .modules.data import MercutoDataService
from .modules.fatigue import MercutoFatigueService
from .modules.identity import MercutoIdentityService
from .modules.media import MercutoMediaService

logger = logging.getLogger(__name__)


class _ModuleBase(Protocol):
    def __init__(self, client: 'MercutoClient', *args: Any, **kwargs: Any) -> None:
        pass


_T = TypeVar('_T', bound=_ModuleBase)


class MercutoClient:
    def __init__(self, url: Optional[str] = None, verify_ssl: bool = True, active_session: Optional[requests.Session] = None) -> None:
        if url is None:
            url = os.environ.get('MERCUTO_API_URL', 'https://api.rockfieldcloud.com.au')
        assert isinstance(url, str)

        if url.endswith('/'):
            url = url[:-1]

        if not url.startswith('https://'):
            raise ValueError(f'Url must be https, is {url}')

        self._url = url
        self.verify_ssl = verify_ssl

        if active_session is None:
            self._current_session = requests.Session()
        else:
            self._current_session = active_session

        self._auth_method: Optional[IAuthenticationMethod] = None
        self._cookies = requests.cookies.RequestsCookieJar()

        self._modules: dict[str, _ModuleBase] = {}

    def url(self) -> str:
        return self._url

    def credentials_key(self) -> str:
        """
        Generate a unique key that identifies the current credentials set.
        """
        if self._auth_method is None:
            raise MercutoClientException("No credentials set")
        return self._auth_method.unique_key()

    def setverify_ssl(self, verify_ssl: bool) -> None:
        self.verify_ssl = verify_ssl

    def copy(self) -> 'MercutoClient':
        return MercutoClient(self._url, self.verify_ssl, self._current_session)

    @contextlib.contextmanager
    def as_credentials(self, api_key: Optional[str] = None,
                       service_token: Optional[str] = None,
                       bearer_token: Optional[str] = None,
                       headers: Optional[Mapping[str, str]] = None) -> Iterator['MercutoClient']:
        """
        Same as .connect(), but as a context manager. Will automatically logout when exiting the context.
        """
        # TODO: We are passing the current session along to re-use connections for speed. Will this cause security issues?
        other = MercutoClient(self._url, self.verify_ssl, self._current_session)
        try:
            yield other.connect(api_key=api_key, service_token=service_token, bearer_token=bearer_token, headers=headers)
        finally:
            other.logout()

    def connect(self, *, api_key: Optional[str] = None,
                service_token: Optional[str] = None,
                bearer_token: Optional[str] = None,
                headers: Optional[Mapping[str, str]] = None) -> 'MercutoClient':
        """
        Attempt to connect using any available method.
        if api_key is provided, use the api_key.
        if service_token is provided, use the service_token.
        if headers is provided, attempt to extract either api_key or service_token from given header set.
            headers should be a dictionary of headers that would be sent in a request. Useful for using existing authenation mechanism for forwarding.

        """
        authentication = create_authentication_method(api_key=api_key, service_token=service_token, bearer_token=bearer_token, headers=headers)
        self.login(authentication)
        return self

    def _update_headers(self, headers: dict[str, str]) -> dict[str, str]:
        base: dict[str, str] = {}

        if self._auth_method is not None:
            self._auth_method.update_header(base)
        base.update(headers)
        return base

    def session(self) -> requests.Session:
        return self._current_session

    def request(self, url: str, method: Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD', 'OPTIONS'],
                params: Optional[dict[str, Any]] = None,
                json: Optional[dict[str, Any]] = None,
                raise_for_status: bool = True,
                **kwargs: Any) -> requests.Response:
        """
        Make an HTTP request to the Mercuto API.
        :param url: The URL path (relative to the base API URL) to make the request to.
        :param method: The HTTP method to use (e.g., 'GET', 'POST', etc.).
        :param params: Optional dictionary of query parameters to include in the request.
        :param json: Optional dictionary to send as a JSON payload in the request body.
        :param raise_for_status: Whether to raise an exception for HTTP error responses.
        :param kwargs: Additional keyword arguments to pass to the requests method.
        :return: The HTTP response object.
        """
        return self._http_request(url, method, params=params, json=json, raise_for_status=raise_for_status, **kwargs)

    def _http_request(self, url: str, method: str,
                      params: Optional[dict[str, Any]] = None,
                      json: Optional[dict[str, Any]] = None,
                      raise_for_status: bool = True,
                      **kwargs: Any) -> requests.Response:
        if url.startswith('/'):
            url = url[1:]
        full_url = f"{self._url}/{url}"

        if 'timeout' not in kwargs:
            kwargs['timeout'] = 10
        kwargs['headers'] = self._update_headers(kwargs.get('headers', {}))

        if 'verify' not in kwargs:
            kwargs['verify'] = self.verify_ssl

        if 'cookies' not in kwargs:
            kwargs['cookies'] = self._cookies

        # Custom parsing json to support NAN
        if json is not None and kwargs.get('data') is None:
            kwargs['data'] = json_stdlib.dumps(json, allow_nan=True)
            kwargs['headers']['Content-Type'] = 'application/json'
            json = None

        start = time.time()
        resp = self._current_session.request(method, full_url, params=params, json=json, **kwargs)
        duration = time.time() - start
        logger.debug("Made request to %s %s in %.2f seconds (code=%s)", method, full_url, duration, resp.status_code)
        if raise_for_status and not resp.ok:
            try:
                error_json = resp.json()
            except Exception:
                raise MercutoHTTPException(resp.text, resp.status_code)
            else:
                if 'detail' in error_json and isinstance(error_json['detail'], str):
                    raise MercutoHTTPException(error_json['detail'], resp.status_code)
                else:
                    raise MercutoHTTPException(resp.text, resp.status_code)
        resp.cookies.update(self._cookies)
        return resp

    def _add_and_fetch_module(self, name: str, module: Type[_T]) -> _T:
        if name not in self._modules:
            self._modules[name] = module(self)
        return self._modules[name]  # type: ignore

    def identity(self) -> 'MercutoIdentityService':
        return self._add_and_fetch_module('identity', MercutoIdentityService)

    def fatigue(self) -> 'MercutoFatigueService':
        return self._add_and_fetch_module('fatigue', MercutoFatigueService)

    def data(self) -> 'MercutoDataService':
        return self._add_and_fetch_module('data', MercutoDataService)

    def core(self) -> 'MercutoCoreService':
        return self._add_and_fetch_module('core', MercutoCoreService)

    def media(self) -> 'MercutoMediaService':
        return self._add_and_fetch_module('media', MercutoMediaService)

    def login(self, authentication: IAuthenticationMethod) -> None:
        self._auth_method = authentication

    def logout(self) -> None:
        self._auth_method = None

    def is_logged_in(self) -> bool:
        return self._auth_method is not None



import os
from datetime import timedelta
from typing import TYPE_CHECKING, Literal, Optional

from pydantic import TypeAdapter

if TYPE_CHECKING:
    from ..client import MercutoClient

from datetime import datetime

from ..exceptions import MercutoHTTPException
from . import PayloadType
from ._util import BaseModel


class Image(BaseModel):
    code: str
    project: str
    mime_type: str
    size_bytes: int
    name: str
    access_url: str
    access_expires: datetime
    camera: Optional[str] = None
    timestamp: Optional[datetime] = None
    event: Optional[str] = None


class Video(BaseModel):
    code: str
    project: str
    start_time: datetime
    end_time: datetime

    mime_type: str
    size_bytes: int
    name: str

    access_url: str
    access_expires: datetime

    event: Optional[str] = None
    camera: Optional[str] = None


class _VideoUploadInitializeResponse(BaseModel):
    request_id: str
    presigned_put_url: str
    presigned_url_expires: datetime


class Healthcheck(BaseModel):
    status: str


class CameraTrigger(BaseModel):
    trigger_channel: str
    pre_interval: timedelta = timedelta(seconds=10)
    post_interval: timedelta = timedelta(seconds=10)
    enabled: bool = True
    trigger_greater_than: Optional[float] = None
    trigger_less_than: Optional[float] = None


class Camera(BaseModel):
    project: str
    label: str
    triggers: list[CameraTrigger]
    camera_type: Literal['BOSCH', 'DIRECT_RTSP', 'STATIC', 'ROCKFIELD-CAMERA-SERVER-VERSION-2']
    encode_timestamp: bool = True
    encode_blur: bool = False
    blur_steps: Optional[int] = None
    blur_sigma: Optional[float] = None
    tunnel_address: Optional[str] = None
    tunnel_port: Optional[int] = None
    tunnel_username: Optional[str] = None
    tunnel_password: Optional[str] = None
    tunnel_key: Optional[str] = None
    camera_ip: Optional[str] = None
    camera_port: Optional[int] = None
    camera_username: Optional[str] = None
    camera_password: Optional[str] = None
    camera_serial: Optional[str] = None
    rtsp_url: Optional[str] = None


# --- TypeAdapters for lists ---
_ImagelistAdapter = TypeAdapter(list[Image])
_VideolistAdapter = TypeAdapter(list[Video])


class MercutoMediaService:
    def __init__(self, client: 'MercutoClient', path: str = '/media') -> None:
        self._client = client
        self._path = path

    def healthcheck(self) -> Healthcheck:
        r = self._client.request(f"{self._path}/healthcheck", "GET")
        return Healthcheck.model_validate_json(r.text)

    # --- Images ---

    def list_images(self, project: str,
                    camera: Optional[str] = None,
                    event: Optional[str] = None,
                    start_time: Optional[datetime] = None,
                    end_time: Optional[datetime] = None,
                    limit: int = 10,
                    offset: int = 0,
                    ascending: bool = True) -> list[Image]:
        params: PayloadType = {
            'project': project,
            'limit': limit,
            'offset': offset,
            'ascending': ascending
        }
        if camera is not None:
            params["camera"] = camera
        if event is not None:
            params["event"] = event
        if start_time is not None:
            if start_time.tzinfo is None:
                raise ValueError("start_time must be timezone-aware")
            params["start_time"] = start_time.isoformat()
        if end_time is not None:
            if end_time.tzinfo is None:
                raise ValueError("end_time must be timezone-aware")
            params["end_time"] = end_time.isoformat()
        r = self._client.request(f"{self._path}/images", "GET", params=params)
        return _ImagelistAdapter.validate_json(r.text)

    def get_image(self, image_code: str) -> Image:
        r = self._client.request(f"{self._path}/images/{image_code}", "GET")
        return Image.model_validate_json(r.text)

    def delete_image(self, image_code: str) -> None:
        self._client.request(f"{self._path}/images/{image_code}", "DELETE")

    def upload_image(self, filename: str, project: str,
                     camera: Optional[str] = None,
                     timestamp: Optional[datetime] = None,
                     event: Optional[str] = None,
                     filedata: Optional[bytes] = None) -> Image:
        """
        Upload an image to the media service.
        Provide either filename (path to file) or filedata (bytes of file) + filename (reference only).
        """
        params: PayloadType = {
            'project': project
        }
        if camera is not None:
            params["camera"] = camera
        if timestamp is not None:
            if timestamp.tzinfo is None:
                raise ValueError("timestamp must be timezone-aware")
            params["timestamp"] = timestamp.isoformat()
        if event is not None:
            params["event"] = event

        if filedata is not None:
            from io import BytesIO
            r = self._client.request(f"{self._path}/images", "PUT", params=params, files={'file': (filename, BytesIO(filedata))})
            return Image.model_validate_json(r.text)
        else:
            with open(filename, 'rb') as f:
                r = self._client.request(f"{self._path}/images", "PUT", params=params, files={'file': f})
        return Image.model_validate_json(r.text)

    # --- Videos ---

    def list_videos(self, project: str,
                    camera: Optional[str] = None,
                    event: Optional[str] = None,
                    start_time: Optional[datetime] = None,
                    end_time: Optional[datetime] = None,
                    limit: int = 10,
                    offset: int = 0,
                    ascending: bool = True) -> list[Video]:
        params: PayloadType = {
            'project': project,
            'limit': limit,
            'offset': offset,
            'ascending': ascending
        }
        if camera is not None:
            params["camera"] = camera
        if event is not None:
            params["event"] = event
        if start_time is not None:
            if start_time.tzinfo is None:
                raise ValueError("start_time must be timezone-aware")
            params["start_time"] = start_time.isoformat()
        if end_time is not None:
            if end_time.tzinfo is None:
                raise ValueError("end_time must be timezone-aware")
            params["end_time"] = end_time.isoformat()
        r = self._client.request(f"{self._path}/videos", "GET", params=params)
        return _VideolistAdapter.validate_json(r.text)

    def get_video(self, video_code: str) -> Video:
        r = self._client.request(f"{self._path}/videos/{video_code}", "GET")
        return Video.model_validate_json(r.text)

    def upload_video(self, filename: str, project: str,
                     start_time: datetime,
                     end_time: datetime,
                     camera: Optional[str] = None,
                     event: Optional[str] = None) -> str:
        """
        Upload a video file to the media service.
        This is a multi-step process.
        First, the request is initialized to get an upload URL.
        Then, the file is uploaded to the provided URL.
        Finally, the upload is finalized.

        :returns: A request ID used to track the status of the uploaded video processing. Pass this to the /requests/{request_id} endpoint
        to check the status and get the video_id once processing is complete.
        """
        import mimetypes
        mime_type = mimetypes.guess_type(filename, strict=False)[0]
        if mime_type is None:
            raise ValueError(f"Could not determine MIME type for file: {filename}")
        if mime_type not in {'video/mp4', 'video/avi', 'video/mov', 'video/mkv'}:
            raise ValueError(f"Unsupported video MIME type: {mime_type}")

        # 1. Make the initialize upload request
        init_payload: PayloadType = {
            'start_time': start_time.isoformat(),
            'end_time': end_time.isoformat(),
            'mime_type': mime_type,
            'filename': os.path.basename(filename)
        }
        if camera is not None:
            init_payload['camera'] = camera
        if event is not None:
            init_payload['event'] = event
        init_request = self._client.request(f"{self._path}/videos", "POST", json=init_payload,
                                            params={'project': project, 'action': 'initialize'})
        init_request_response = _VideoUploadInitializeResponse.model_validate_json(init_request.text)

        # 2. Upload the video file
        with open(filename, 'rb') as f:
            files = {'file': f}
            resp = self._client.session().put(init_request_response.presigned_put_url, files=files, verify=self._client.verify_ssl)
            if not resp.ok:
                raise MercutoHTTPException(f"Video upload failed: {resp.text}", resp.status_code)

        # 3. Finalize the upload
        self._client.request(f"{self._path}/videos", "POST", params={
            'project': project,
            'action': 'commit',
            'request_id': init_request_response.request_id
        })

        return init_request_response.request_id

    # --- Cameras ---

    def create_camera(self, project: str,
                      label: str,
                      triggers: list[CameraTrigger],
                      encode_timestamp: bool = True,
                      encode_blur: bool = False,
                      blur_steps: Optional[int] = None,
                      blur_sigma: Optional[float] = None,
                      tunnel_address: Optional[str] = None,
                      tunnel_port: Optional[int] = None,
                      tunnel_username: Optional[str] = None,
                      tunnel_password: Optional[str] = None,
                      tunnel_key: Optional[str] = None,
                      camera_ip: Optional[str] = None,
                      camera_port: Optional[int] = None,
                      camera_username: Optional[str] = None,
                      camera_password: Optional[str] = None,
                      camera_serial: Optional[str] = None,
                      camera_type: Literal['BOSCH', 'DIRECT_RTSP', 'STATIC', 'ROCKFIELD-CAMERA-SERVER-VERSION-2'] = 'DIRECT_RTSP',
                      rtsp_url: Optional[str] = None,) -> Camera:
        payload: PayloadType = {
            'label': label,
            'encode_timestamp': encode_timestamp,
            'encode_blur': encode_blur,
        }
        if blur_steps is not None:
            payload['blur_steps'] = blur_steps
        if blur_sigma is not None:
            payload['blur_sigma'] = blur_sigma
        if tunnel_address is not None:
            tunnel_payload: PayloadType = {
                'address': tunnel_address,
                'port': tunnel_port,
                'username': tunnel_username,
                'password': tunnel_password,
                'key': tunnel_key
            }
            payload['ssh_tunnel'] = tunnel_payload
        if camera_ip is not None:
            payload['camera_ip'] = camera_ip
        if camera_port is not None:
            payload['camera_port'] = camera_port
        if camera_username is not None:
            payload['camera_username'] = camera_username
        if camera_password is not None:
            payload['camera_password'] = camera_password
        if camera_serial is not None:
            payload['camera_serial'] = camera_serial
        payload['camera_type'] = camera_type
        if rtsp_url is not None:
            payload['rtsp_url'] = rtsp_url

        if triggers:
            payload['triggers'] = [trigger.model_dump(mode='json') for trigger in triggers]  # type: ignore

        r = self._client.request(f"{self._path}/cameras", "PUT", json=payload, params={'project': project})
        return Camera.model_validate_json(r.text)



import fnmatch
import logging
import os
from typing import Optional

from .. import MercutoClient, MercutoHTTPException
from ..modules.core import Project
from ..modules.data import (Channel, ChannelClassification, Datatable,
                            SecondaryDataSample)
from ..util import batched, get_my_public_ip
from .parsers import detect_parser

logger = logging.getLogger(__name__)

NON_RETRYABLE_ERRORS = {400, 404, 409}  # HTTP status codes that indicate non-retryable errors


class MercutoIngester:
    def __init__(self, project_code: str, api_key: str, hostname: str = 'https://api.rockfieldcloud.com.au') -> None:
        self._client = MercutoClient(url=hostname)
        self._api_key = api_key
        self._project_code = project_code

        self._project: Optional[Project] = None
        self._secondary_channels: Optional[list[Channel]] = None
        self._datatables: Optional[list[Datatable]] = None

        self._channel_map: dict[str, str] = {}

    def _refresh_mercuto_data(self) -> None:
        with self._client.as_credentials(api_key=self._api_key) as client:
            self._project = client.core().get_project(self._project_code)
            assert self._project.code == self._project_code

            self._secondary_channels = client.data().list_channels(self._project_code, classification=ChannelClassification.SECONDARY)
            self._datatables = client.data().list_datatables(self._project_code)

        self._channel_map.update({c.label: c.code for c in self._secondary_channels})

    def _can_process(self) -> bool:
        return self._project is not None and self._secondary_channels is not None and self._datatables is not None

    def update_mapping(self, mapping: dict[str, str]) -> None:
        """
        Update the channel label to channel code mapping.
        """
        self._channel_map.update(mapping)
        logger.info(f"Updated channel mapping: {self._channel_map}")

    @property
    def project_code(self) -> str:
        return self._project_code

    def ping(self) -> None:
        """
        Ping the Mercuto serverto update the last seen IP address.
        """
        ip = get_my_public_ip()
        with self._client.as_credentials(api_key=self._api_key) as client:
            client.core().ping_project(self.project_code, ip_address=ip)
            logging.info(f"Pinged Mercuto server from IP: {ip} for project: {self.project_code}")

    def matching_datatable(self, filename: str) -> str | None:
        """
        Check if any datatables on the project match this file name.
        Returns the datatable code if a match is found, otherwise None.
        """
        if self._datatables is None:
            raise ValueError("Datatables not loaded. Call _refresh_mercuto_data() first.")

        basename = os.path.basename(filename)

        def matches(test: str) -> bool:
            """
            test should be a pattern or a filename.
            E.g. "my_data.csv" or "my_data*.csv", or "/path/to/my_data*.csv"
            Do wildcard matching as well as prefix matching.
            """
            test_base = os.path.basename(test)
            if fnmatch.fnmatch(basename, test_base):
                return True
            lhs, _ = os.path.splitext(test_base)
            if basename.startswith(lhs):
                return True
            return False

        for dt in self._datatables:
            # Match using datatable pattern
            if matches(dt.name):
                return dt.code
        return None

    def _upload_samples(self, samples: list[SecondaryDataSample]) -> bool:
        """
        Upload samples to the Mercuto project.
        """
        try:
            with self._client.as_credentials(api_key=self._api_key) as client:
                for batch in batched(samples, 500):
                    client.data().insert_secondary_samples(self.project_code, batch)
            return True
        except MercutoHTTPException as e:
            if e.status_code in NON_RETRYABLE_ERRORS:
                logger.exception(
                    "Error indicates bad file that should not be retried. Skipping.")
                return True
            else:
                return False

    def _upload_file(self, file_path: str, datatable_code: str) -> bool:
        """
        Upload a file to the Mercuto project.
        """
        logging.info(f"Uploadeding file {file_path} to datatable {datatable_code} in project {self.project_code}")
        try:
            with self._client.as_credentials(api_key=self._api_key) as client:
                client.data().upload_file(
                    project=self.project_code,
                    datatable=datatable_code,
                    file=file_path,
                )
            return True
        except MercutoHTTPException as e:
            if e.status_code in NON_RETRYABLE_ERRORS:
                logger.exception(
                    "Error indicates bad file that should not be retried. Skipping.")
                return True
            else:
                return False

    def process_file(self, file_path: str) -> bool:
        """
        Process the received file.
        """

        if not self._can_process():
            logging.info("Refreshing Mercuto data...")
            self._refresh_mercuto_data()
            if not self._can_process():
                logging.error("Failed to refresh Mercuto data. Cannot process file yet.")
                return False

        logging.info(f"Processing file: {file_path}")
        datatable_code = self.matching_datatable(file_path)
        if datatable_code:
            logger.info(f"Matched datatable code: {datatable_code} for file: {file_path}")
            return self._upload_file(file_path, datatable_code)
        else:
            parser = detect_parser(file_path)
            samples = parser(file_path, self._channel_map)
            if not samples:
                logging.warning(f"No samples found in file: {file_path}")
                return True
            return self._upload_samples(samples)



import itertools
import shutil
from datetime import timedelta
from pathlib import Path
from typing import Iterable, Iterator, TypeVar

import requests


def timedelta_isoformat(td: timedelta) -> str:
    """
    ISO 8601 encoding for Python timedelta object.
    Taken from pydantic source:
        https://github.com/pydantic/pydantic/blob/3704eccce4661455acdda1cdcf716bd4b3382e08/pydantic/deprecated/json.py#L135-L140

    """
    minutes, seconds = divmod(td.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return f'{"-" if td.days < 0 else ""}P{abs(td.days)}DT{hours:d}H{minutes:d}M{seconds:d}.{td.microseconds:06d}S'


def get_my_public_ip() -> str:
    """
    Fetches the public IP address of the machine making the request.
    Uses the 'checkip.amazonaws.com' service to retrieve the IP address.
    :return The public IP address as a string in the form 'x.x.x.x'.
    :raises
        requests.RequestException: If the request to the IP service fails.
        requests.Timeout: If the request times out.
    """
    r = requests.get('https://checkip.amazonaws.com', timeout=30)
    r.raise_for_status()
    return r.content.decode().strip()


def get_directory_size(directory: str) -> int:
    """
    Returns the total size (in bytes) of the target directory, including all subdirectories.

    :param directory: Path to the target directory.
    :return: Total size in bytes.
    """
    dir_path = Path(directory)
    return sum(f.stat().st_size for f in dir_path.rglob("*") if f.is_file())


def get_free_space_excluding_files(directory: str) -> int:
    """
    Returns the number of free bytes on the partition of the target directory,
    excluding the total size of files in that directory.

    :param directory: Path to the target directory.
    :return: Free bytes available in the partition after subtracting file sizes.
    """
    # Get partition's free space
    total, used, free = shutil.disk_usage(directory)

    # Calculate the total size of files in the directory
    files_size = get_directory_size(directory)

    # Exclude file sizes from free space
    return max(0, free - files_size)


T = TypeVar('T')

try:
    from itertools import batched  # type: ignore
except ImportError:
    # Python < 3.12
    def batched(iterable: Iterable[T], n: int) -> Iterator[tuple[T, ...]]:  # type: ignore[no-redef]
        """
        Implementation of itertools.batched for < Python 3.12
        """
        it = iter(iterable)
        while True:
            chunk = tuple(itertools.islice(it, n))
            if not chunk:
                break
            yield chunk



from typing import Optional, Protocol

import pytz

from ...modules.data import SecondaryDataSample
from .campbell import parse_campbell_file
from .worldsensing import (parse_worldsensing_compact_file,
                           parse_worldsensing_standard_file)


class Parser(Protocol):
    def __call__(self, filename: str, label_to_channel_code: dict[str, str],
                 timezone: Optional[pytz.BaseTzInfo] = None) -> list[SecondaryDataSample]:
        """
        Parse the file and return a list of SecondaryDataSample objects.
        """
        pass


def detect_parser(filename: str) -> Parser:
    """
    Detect the type of the file based on its content.
    Returns one of: "worldsensing_compact", "worldsensing_standard", "campbell", or raises ValueError if unknown.
    """
    with open(filename, 'r') as f:
        first_line = f.readline().strip()
        if first_line.startswith('"TOA5",'):
            return parse_campbell_file
        elif first_line.startswith('"Datalogger","compacted"'):
            return parse_worldsensing_compact_file
        elif first_line.startswith('"Node ID",'):
            return parse_worldsensing_standard_file
        else:
            raise ValueError(f"Unknown file type for {filename}")


__all__ = [
    "parse_campbell_file",
    "parse_worldsensing_standard_file",
    "parse_worldsensing_compact_file",
    "detect_parser",
]



import logging
from datetime import datetime
from typing import Literal, Optional

from ..client import MercutoClient
from ..modules.fatigue import MercutoFatigueService
from ._utility import EnforceOverridesMeta

logger = logging.getLogger(__name__)


class MockMercutoFatigueService(MercutoFatigueService, metaclass=EnforceOverridesMeta):
    def __init__(self, client: 'MercutoClient'):
        super().__init__(client=client, path='/mock-fatigue-service-method-not-implemented')

    def delete_cycle_counts(
        self, project: str, start_time: datetime, end_time: datetime, ignore_if_not_configured: bool = False
    ) -> None:
        pass

    def calculate_cycle_counts(
        self,
        project: str,
        event: str,
        presigned_url: str,
        mime_type: Literal['application/feather'],
        url_expiry: Optional[datetime] = None,
        ignore_if_not_configured: bool = False
    ) -> None:
        pass



import logging
from datetime import datetime
from typing import Optional

import pytz
from dateutil import parser

from ...modules.data import SecondaryDataSample

logger = logging.getLogger(__name__)


def _clean(s: str) -> str:
    s = s.strip()
    if s.startswith('"') and s.endswith('"'):
        s = s[1:-1]
    return s


def _clean_number(s: str) -> float | None:
    cleaned = _clean(s)
    if cleaned == 'NAN' or cleaned == 'N/A' or cleaned == 'NaN' or cleaned == 'nan' or cleaned == 'Nan':
        return float('nan')
    try:
        return float(cleaned)
    except ValueError:
        return None


def _parse_header(header: str) -> list[str]:
    columns = [h.strip() for h in header.strip().split(",")]
    if columns[0] not in ('"TIMESTAMP"', 'timestamp', 'TIMESTAMP', '"Date-and-time"'):
        raise ValueError(
            f"Invalid header found: {columns[0]}, expecting TIMESTAMP.")
    # Columns have quotes around them, remove them
    return [_clean(c) for c in columns[1:]]


def _parse_csv_line(line: str, sep: str = ',', timestamp_index: int = 0) -> tuple[datetime, list[float | None]]:
    """
    Returns timestamp, values
    """
    values = line.strip().split(sep)
    if len(values) < 2:
        raise ValueError(f"Invalid number of values found: {len(values)}")
    # First value is timestamp
    timestamp = _clean(values[timestamp_index])
    try:
        dt = parser.parse(timestamp)
    except ValueError as e:
        raise ValueError(
            f"Failed to parse timestamp: {timestamp} - {e}") from e
    # Rest are values
    return dt, [_clean_number(v) for v in values[timestamp_index+1:]]


def parse_generic_csv_file(filename: str, label_to_channel_code: dict[str, str],
                           header_index: int, data_index: int,
                           timezone: Optional[pytz.BaseTzInfo] = None) -> list[SecondaryDataSample]:
    """
    header index: Number of lines to skip before header
    data index: Number of lines to skip after the header before data

    We are avoiding using pandas here to keep dependencies minimal as this is often run on edge devices.
    """

    output: list[SecondaryDataSample] = []
    with open(filename, "r") as f:
        for _ in range(header_index):
            next(f, None)
        header = next(f, None)
        if header is None:
            logging.error(f"Failed to read header from file: {filename}")
            return []
        try:
            header_columns = _parse_header(header)
        except ValueError as e:
            logging.error(f"Failed to parse header: {e}")
            return []

        # Next 2 lines are metadata, skip
        for _ in range(data_index):
            next(f, None)
        while (line := next(f, None)):
            try:
                timestamp, line_values = _parse_csv_line(line)
            except ValueError as e:
                logging.error(f"Failed to parse line: {e}")
                return []

            if len(header_columns) != len(line_values):
                logging.error(
                    f"Invalid number of values found: {len(line_values)}, expected {len(header_columns)}")
                return []

            if timezone is not None and timestamp.tzinfo is None:
                timestamp = timezone.localize(timestamp)

            for header, value in zip(header_columns, line_values):
                if header not in label_to_channel_code:
                    logger.error(f"Label not found in table map: {header}")
                    continue

                if value is None:
                    logger.error(
                        f"Failed to parse value: {value} for column {header}")
                    continue
                channel_code = label_to_channel_code[header]

                logger.debug(
                    f"Adding entry for label: {header} with value: {value} and timestamp: {timestamp}")
                output.append(SecondaryDataSample(timestamp=timestamp,
                              channel=channel_code, value=value))
    return output



from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Literal, Optional

from pydantic import TypeAdapter

from . import _PayloadType
from ._util import BaseModel, serialise_timedelta

if TYPE_CHECKING:
    from ..client import MercutoClient


class ProjectStatus(BaseModel):
    last_ping: Optional[str]
    ip_address: Optional[str]


class Project(BaseModel):
    code: str
    name: str
    project_number: str
    active: bool
    description: str
    latitude: Optional[float]
    longitude: Optional[float]
    timezone: str
    display_timezone: Optional[str]
    tenant: str
    status: ProjectStatus
    commission_date: datetime


class WidgetConfig(BaseModel):
    type: str
    config: dict[Any, Any]


class WidgetColumn(BaseModel):
    size: Optional[str | int]
    widget: WidgetConfig


class WidgetRow(BaseModel):
    columns: list[WidgetColumn]
    height: int
    title: str
    breakpoint: Optional[str]


class Dashboard(BaseModel):
    icon: Optional[str]
    name: Optional[str]
    banner_image: Optional[str]
    widgets: Optional[list[WidgetRow]]
    fullscreen: Optional[bool]


class Dashboards(BaseModel):
    dashboards: list[Dashboard]


class ProjectEventDetection(BaseModel):
    enabled: bool
    datatables: list[str]
    max_duration: timedelta
    max_files: int
    maximise: bool
    overlap_period: timedelta
    split_interval_cron: Optional[str]


class ItemCode(BaseModel):
    code: str


class EventTag(BaseModel):
    tag: str
    value: Any | None


class Object(BaseModel):
    code: str
    mime_type: str
    size_bytes: int
    name: str
    event: ItemCode | None
    project: ItemCode
    access_url: str | None
    access_expires: datetime | None


class Event(BaseModel):
    code: str
    project: ItemCode
    start_time: datetime
    end_time: datetime
    objects: list[Object]
    tags: list[EventTag]


UserContactMethod = Literal['EMAIL', 'SMS']


class ContactGroup(BaseModel):
    project: str
    code: str
    label: str
    users: dict[str, list[UserContactMethod]]


class Condition(BaseModel):
    code: str
    source: str
    description: str
    upper_exclusive_bound: Optional[float]
    lower_inclusive_bound: Optional[float]
    neutral_position: float


class AlertConfiguration(BaseModel):
    code: str
    project: str
    label: str
    conditions: list[Condition]
    contact_group: Optional[ContactGroup]
    retrigger_interval: Optional[datetime]


class AlertLogConditionEntry(BaseModel):
    condition: Condition
    start_value: float
    start_time: str
    start_percentile: float

    peak_value: float
    peak_time: str
    peak_percentile: float

    end_value: float
    end_time: str
    end_percentile: float


class AlertLogComment(BaseModel):
    user_code: str
    comment: str
    created_at: str


class AlertLog(BaseModel):
    code: str
    project: str
    event: Optional[str]
    acknowledged: bool
    fired_at: str
    configuration: str
    conditions: list[AlertLogConditionEntry]
    comments: list[AlertLogComment]


class AlertSummary(BaseModel):
    alerts: list[AlertLog]
    total: int


class Healthcheck(BaseModel):
    ephemeral_warehouse: str
    ephemeral_document_store: str
    cache: str
    database: str


class DeviceType(BaseModel):
    code: str
    description: str
    manufacturer: str
    model_number: str


class DeviceChannel(BaseModel):
    channel: str
    field: str


class Device(BaseModel):
    code: str
    project: ItemCode
    label: str
    location_description: Optional[str]
    device_type: DeviceType
    groups: list[str]
    channels: list[DeviceChannel]


class EventAggregate(BaseModel):
    aggregate: Literal["max", "greatest", "min", "median",
                       "abs-max", "mean", "rms", "peak-to-peak", "daf"]
    enabled: bool = True
    options: Optional[dict[str, Any]] = None


class ScheduledReport(BaseModel):
    code: str
    project: str
    label: str
    revision: str
    schedule: Optional[str]
    contact_group: Optional[str]
    last_scheduled: Optional[str]


class ScheduledReportLog(BaseModel):
    code: str
    report: str
    scheduled_start: Optional[str]
    actual_start: str
    actual_finish: Optional[str]
    status: Literal['IN_PROGRESS', 'COMPLETED', 'FAILED']
    message: Optional[str]
    access_url: Optional[str]
    mime_type: Optional[str]
    filename: Optional[str]


class ReportSourceCodeRevision(BaseModel):
    code: str
    revision_date: datetime
    description: str
    source_code_url: str


_ProjectListAdapter = TypeAdapter(list[Project])
_EventsListAdapter = TypeAdapter(list[Event])
_DevicesListAdapter = TypeAdapter(list[Device])
_DeviceTypeListAdapter = TypeAdapter(list[DeviceType])
_ContactGroupListAdapter = TypeAdapter(list[ContactGroup])
_ScheduledReportListAdapter = TypeAdapter(list[ScheduledReport])
_ScheduledReportLogListAdapter = TypeAdapter(list[ScheduledReportLog])


class MercutoCoreService:
    def __init__(self, client: 'MercutoClient') -> None:
        self._client = client

    def healthcheck(self) -> Healthcheck:
        r = self._client._http_request("/healthcheck", "GET")
        return Healthcheck.model_validate_json(r.text)

    # Projects

    def get_project(self, code: str) -> Project:
        if len(code) == 0:
            raise ValueError("Project code must not be empty")
        r = self._client._http_request(f'/projects/{code}', 'GET')
        return Project.model_validate_json(r.text)

    def list_projects(self) -> list[Project]:
        r = self._client._http_request('/projects', 'GET')
        return _ProjectListAdapter.validate_json(r.text)

    def create_project(self, name: str, project_number: str, description: str, tenant: str,
                       timezone: str,
                       latitude: Optional[float] = None,
                       longitude: Optional[float] = None) -> Project:

        payload: _PayloadType = {
            'name': name,
            'project_number': project_number,
            'description': description,
            'tenant_code': tenant,
            'timezone': timezone,
        }
        if latitude is not None:
            payload['latitude'] = latitude
        if longitude is not None:
            payload['longitude'] = longitude

        r = self._client._http_request('/projects', 'PUT', json=payload)
        return Project.model_validate_json(r.text)

    def ping_project(self, project: str, ip_address: str) -> None:
        self._client._http_request(
            f'/projects/{project}/ping', 'POST', json={'ip_address': ip_address})

    def create_dashboard(self, project_code: str, dashboards: Dashboards) -> None:
        json = dashboards.model_dump()
        self._client._http_request(
            f'/projects/{project_code}/dashboard', 'POST', json=json)

    def set_project_event_detection(self, project: str, datatables: list[str]) -> ProjectEventDetection:
        if len(datatables) == 0:
            raise ValueError(
                'At least one datatable must be provided to enable event detection')

        params: _PayloadType = {
            "enabled": True,
            "datatables": datatables
        }
        r = self._client._http_request(
            f'/projects/{project}/event-detection', 'POST', json=params)
        return ProjectEventDetection.model_validate_json(r.text)

    # EVENTS

    def create_event(self, project: str, start_time: datetime, end_time: datetime) -> Event:
        if start_time.tzinfo is None or end_time.tzinfo is None:
            raise ValueError("Timestamp must be timezone aware")

        json: _PayloadType = {
            'project': project,
            'start_time': start_time.isoformat(),
            'end_time': end_time.isoformat(),
        }
        r = self._client._http_request('/events', 'PUT', json=json)
        return Event.model_validate_json(r.text)

    def list_events(self, project: str) -> list[Event]:
        params: _PayloadType = {'project_code': project}
        r = self._client._http_request('/events', 'GET', params=params)
        return _EventsListAdapter.validate_json(r.text)

    def get_event(self, event: str) -> Event:
        r = self._client._http_request(f'/events/{event}', 'GET')
        return Event.model_validate_json(r.text)

    def delete_event(self, event: str) -> None:
        self._client._http_request(f'/events/{event}', 'DELETE')

    def get_nearest_event(
        self,
        project_code: str,
        to: datetime,
        maximum_delta: timedelta | None = None,
    ) -> Event:
        params: _PayloadType = {
            'project_code': project_code,
            'to': to.isoformat(),
        }
        if maximum_delta is not None:
            params['maximum_delta'] = serialise_timedelta(maximum_delta)

        r = self._client._http_request('/events/nearest', 'GET', params=params)
        return Event.model_validate_json(r.text)

    def set_event_aggregates(self, project: str, aggregates: list[EventAggregate]) -> None:
        self._client._http_request('/aggregates', 'PUT',
                                   json=[agg.model_dump(mode='json') for agg in aggregates],  # type: ignore
                                   params={'project_code': project})

    # ALERTS

    def get_condition(self, code: str) -> Condition:
        r = self._client._http_request(f'/alerts/conditions/{code}', 'GET')
        return Condition.model_validate_json(r.text)

    def create_condition(self, source: str, description: str, *,
                         lower_bound: Optional[float] = None,
                         upper_bound: Optional[float] = None,
                         neutral_position: float = 0) -> Condition:
        json: _PayloadType = {
            'source_channel_code': source,
            'description': description,
            'neutral_position': neutral_position
        }
        if lower_bound is not None:
            json['lower_inclusive_bound'] = lower_bound
        if upper_bound is not None:
            json['upper_exclusive_bound'] = upper_bound
        r = self._client._http_request('/alerts/conditions', 'PUT',  json=json)
        return Condition.model_validate_json(r.text)

    def create_alert_configuration(self, label: str,
                                   conditions: list[str],
                                   contact_group: Optional[str] = None) -> AlertConfiguration:
        json: _PayloadType = {
            'label': label,
            'conditions': conditions,

        }
        if contact_group is not None:
            json['contact_group'] = contact_group
        r = self._client._http_request(
            '/alerts/configurations', 'PUT', json=json)
        return AlertConfiguration.model_validate_json(r.text)

    def get_alert_configuration(self, code: str) -> AlertConfiguration:
        r = self._client._http_request(f'/alerts/configurations/{code}', 'GET')
        return AlertConfiguration.model_validate_json(r.text)

    def list_alert_logs(
            self,
            project: str | None = None,
            configuration: str | None = None,
            channels: list[str] | None = None,
            start_time: datetime | str | None = None,
            end_time: datetime | str | None = None,
            limit: int = 10,
            offset: int = 0,
            latest_only: bool = False,
    ) -> AlertSummary:
        params: _PayloadType = {
            'limit': limit,
            'offset': offset,
            'latest_only': latest_only,
        }

        if project is not None:
            params['project'] = project
        if configuration is not None:
            params['configuration_code'] = configuration
        if channels is not None:
            params['channels'] = channels
        if start_time is not None:
            params['start_time'] = start_time.isoformat() if isinstance(
                start_time, datetime) else start_time
        if end_time is not None:
            params['end_time'] = end_time.isoformat() if isinstance(
                end_time, datetime) else end_time

        r = self._client._http_request('/alerts/logs', 'GET', params=params)
        return AlertSummary.model_validate_json(r.text)

    # DEVICES

    def list_device_types(self) -> list[DeviceType]:
        r = self._client._http_request('/devices/types', 'GET')
        return _DeviceTypeListAdapter.validate_json(r.text)

    def create_device_type(self, description: str, manufacturer: str, model_number: str) -> DeviceType:
        json: _PayloadType = {
            'description': description,
            'manufacturer': manufacturer,
            'model_number': model_number
        }
        r = self._client._http_request('/devices/types', 'PUT',  json=json)
        return DeviceType.model_validate_json(r.text)

    def list_devices(self, project_code: str, limit: int, offset: int) -> list[Device]:
        params: _PayloadType = {
            'project_code': project_code,
            'limit': limit,
            'offset': offset
        }
        r = self._client._http_request('/devices', 'GET', params=params)
        return _DevicesListAdapter.validate_json(r.text)

    def get_device(self, device_code: str) -> Device:
        r = self._client._http_request(f'/devices/{device_code}', 'GET')
        return Device.model_validate_json(r.text)

    def create_device(self,
                      project_code: str,
                      label: str,
                      device_type_code: str,
                      groups: list[str],
                      location_description: Optional[str] = None,
                      channels: Optional[list[DeviceChannel]] = None) -> Device:
        json: _PayloadType = {
            'project_code': project_code,
            'label': label,
            'device_type_code': device_type_code,
            'groups': groups,
        }
        if location_description is not None:
            json['location_description'] = location_description
        if channels is not None:
            json['channels'] = [channel.model_dump(mode='json') for channel in channels]  # type: ignore[assignment]
        r = self._client._http_request('/devices', 'PUT', json=json)
        return Device.model_validate_json(r.text)

    # Contacts

    def list_contact_groups(self, project: Optional[str] = None) -> list[ContactGroup]:
        params: _PayloadType = {}
        if project is not None:
            params['project'] = project
        r = self._client._http_request(
            '/notifications/contact_groups', 'GET', params=params)
        return _ContactGroupListAdapter.validate_json(r.text)

    def get_contact_group(self, code: str) -> ContactGroup:
        r = self._client._http_request(
            f'/notifications/contact_groups/{code}', 'GET')
        return ContactGroup.model_validate_json(r.text)

    def create_contact_group(self, project: str, label: str, users: dict[str, list[UserContactMethod]]) -> ContactGroup:
        r = self._client._http_request('/notifications/contact_groups', 'PUT',
                                       json={
                                           'project': project,
                                           'label': label,
                                           'users': users
                                       })
        return ContactGroup.model_validate_json(r.text)

    # Reports
    def list_reports(self, project: Optional[str] = None) -> list['ScheduledReport']:
        params: _PayloadType = {}
        if project is not None:
            params['project'] = project
        r = self._client._http_request(
            '/reports/scheduled', 'GET', params=params)
        return _ScheduledReportListAdapter.validate_json(r.text)

    def create_report(self, project: str, label: str, schedule: str, revision: str,
                      api_key: Optional[str] = None, contact_group: Optional[str] = None) -> ScheduledReport:
        json: _PayloadType = {
            'project': project,
            'label': label,
            'schedule': schedule,
            'revision': revision,
            'execution_role_api_key': api_key,
            'contact_group': contact_group
        }
        r = self._client._http_request('/reports/scheduled', 'PUT', json=json)
        return ScheduledReport.model_validate_json(r.text)

    def generate_report(self, report: str, timestamp: datetime, mark_as_scheduled: bool = False) -> ScheduledReportLog:
        r = self._client._http_request(f'/reports/scheduled/{report}/generate', 'PUT', json={
            'timestamp': timestamp.isoformat(),
            'mark_as_scheduled': mark_as_scheduled
        })
        return ScheduledReportLog.model_validate_json(r.text)

    def list_report_logs(self, report: str, project: Optional[str] = None) -> list[ScheduledReportLog]:
        params: _PayloadType = {}
        if project is not None:
            params['project'] = project
        r = self._client._http_request(
            f'/reports/scheduled/{report}/logs', 'GET', params=params)
        return _ScheduledReportLogListAdapter.validate_json(r.text)

    def get_report_log(self, report: str, log: str) -> ScheduledReportLog:
        r = self._client._http_request(
            f'/reports/scheduled/{report}/logs/{log}', 'GET')
        return ScheduledReportLog.model_validate_json(r.text)

    def create_report_revision(self, project: str, revision_date: datetime, description: str, source_code_data_url: str) -> ReportSourceCodeRevision:
        json = {
            'revision_date': revision_date.isoformat(),
            'description': description,
            'source_code_data_url': source_code_data_url,
        }
        r = self._client._http_request(
            '/reports/revisions', 'PUT', json=json, params={'project': project})
        return ReportSourceCodeRevision.model_validate_json(r.text)



import contextlib
from typing import Callable, Iterator, Optional

from mercuto_client.modules.identity import VerifyMyPermissions

from ..client import MercutoClient


@contextlib.contextmanager
def mock_mercuto(data: bool = True,
                 identity: bool = True,
                 fatigue: bool = True,
                 core: bool = True,
                 media: bool = True,
                 verify_service_token: Optional[Callable[[str], VerifyMyPermissions]] = None) -> Iterator[None]:
    """
    While this context is active, all calls to MercutoClient will use mocked services.

    :param data: Whether to mock the data module.
    :param identity: Whether to mock the identity module.
    :param fatigue: Whether to mock the fatigue module.
    :param verify_service_token: Optional function to mock the verify_service_token behavior. Only used for the mock identity service.
    """
    with contextlib.ExitStack() as stack:
        if data:
            stack.enter_context(mock_data_module())
        if identity:
            stack.enter_context(mock_identity_module(verify_service_token=verify_service_token))
        if fatigue:
            stack.enter_context(mock_fatigue_module())
        if core:
            stack.enter_context(mock_core_module())
        if media:
            stack.enter_context(mock_media_module())
        yield


@contextlib.contextmanager
def mock_core_module() -> Iterator[None]:
    from .mock_core import MockMercutoCoreService
    original = MercutoClient.core

    _cache: Optional[MockMercutoCoreService] = None

    def stub(self: MercutoClient) -> MockMercutoCoreService:
        nonlocal _cache
        if _cache is None:
            _cache = MockMercutoCoreService(self)
        _cache._client = self
        return _cache

    try:
        setattr(MercutoClient, 'core', stub)
        yield
    finally:
        setattr(MercutoClient, 'core', original)


@contextlib.contextmanager
def mock_data_module() -> Iterator[None]:
    from .mock_data import MockMercutoDataService
    original = MercutoClient.data

    _cache: Optional[MockMercutoDataService] = None

    def stub(self: MercutoClient) -> MockMercutoDataService:
        nonlocal _cache
        if _cache is None:
            _cache = MockMercutoDataService(self)
        _cache._client = self
        return _cache

    try:
        setattr(MercutoClient, 'data', stub)
        yield
    finally:
        setattr(MercutoClient, 'data', original)


@contextlib.contextmanager
def mock_identity_module(verify_service_token: Optional[Callable[[str], VerifyMyPermissions]] = None) -> Iterator[None]:
    from .mock_identity import MockMercutoIdentityService
    original = MercutoClient.identity

    _cache: Optional[MockMercutoIdentityService] = None

    def stub(self: MercutoClient) -> MockMercutoIdentityService:
        nonlocal _cache
        if _cache is None:
            _cache = MockMercutoIdentityService(self, verify_service_token=verify_service_token)
        _cache._client = self
        return _cache

    try:
        setattr(MercutoClient, 'identity', stub)
        yield
    finally:
        setattr(MercutoClient, 'identity', original)


@contextlib.contextmanager
def mock_fatigue_module() -> Iterator[None]:
    from .mock_fatigue import MockMercutoFatigueService
    original = MercutoClient.fatigue

    _cache: Optional[MockMercutoFatigueService] = None

    def stub(self: MercutoClient) -> MockMercutoFatigueService:
        nonlocal _cache
        if _cache is None:
            _cache = MockMercutoFatigueService(self)
        _cache._client = self
        return _cache

    try:
        setattr(MercutoClient, 'fatigue', stub)
        yield
    finally:
        setattr(MercutoClient, 'fatigue', original)


@contextlib.contextmanager
def mock_media_module() -> Iterator[None]:
    from .mock_media import MockMercutoMediaService
    original = MercutoClient.media

    _cache: Optional[MockMercutoMediaService] = None

    def stub(self: MercutoClient) -> MockMercutoMediaService:
        nonlocal _cache
        if _cache is None:
            _cache = MockMercutoMediaService(self)
        _cache._client = self
        return _cache

    try:
        setattr(MercutoClient, 'media', stub)
        yield
    finally:
        setattr(MercutoClient, 'media', original)



import logging
import uuid
from datetime import datetime

from ..client import MercutoClient
from ..exceptions import MercutoHTTPException
from ..modules.core import Event, ItemCode, MercutoCoreService
from ._utility import EnforceOverridesMeta

logger = logging.getLogger(__name__)


class MockMercutoCoreService(MercutoCoreService, metaclass=EnforceOverridesMeta):
    def __init__(self, client: 'MercutoClient'):
        super().__init__(client=client)
        self._events: dict[str, Event] = {}

    def create_event(self, project: str, start_time: datetime, end_time: datetime) -> Event:
        event = Event(code=str(uuid.uuid4()), project=ItemCode(code=project), start_time=start_time, end_time=end_time, objects=[], tags=[])
        self._events[event.code] = event
        return event

    def get_event(self, event: str) -> Event:
        if event not in self._events:
            raise MercutoHTTPException(status_code=404, message=f"Event {event} not found")
        return self._events[event]

    def list_events(self, project: str) -> list[Event]:
        return [event for event in self._events.values() if event.project.code == project]



import logging
import uuid
from dataclasses import dataclass
from typing import Callable, Optional

from ..client import MercutoClient
from ..exceptions import MercutoHTTPException
from ..modules.identity import (CurrentUser, HiddenUserAPIKey,
                                MercutoIdentityService, PermissionGroup,
                                Tenant, User, UserDetails, VerifyMyPermissions,
                                VisibleUserAPIKey)
from ._utility import EnforceOverridesMeta

logger = logging.getLogger(__name__)


class MockMercutoIdentityService(MercutoIdentityService, metaclass=EnforceOverridesMeta):
    @dataclass
    class _GeneratedUserApiKey:
        user: str
        api_key: str
        hidden: HiddenUserAPIKey

    def __init__(self, client: 'MercutoClient', verify_service_token: Optional[Callable[[str], VerifyMyPermissions]] = None) -> None:
        """
        Initialize the mock identity service.
        :param client: The MercutoClient instance.
        :param verify_service_token: Optional callable to verify service tokens.
        """

        super().__init__(client=client, path='/mock-identity-service-method-not-implemented')
        self._verify_service_token = verify_service_token

        # In-memory stores for mock data
        self._users: dict[str, User] = {}
        self._user_details: dict[str, UserDetails] = {}
        self._tenants: dict[str, Tenant] = {}
        self._permission_groups: dict[str, PermissionGroup] = {}

        # Maps API Key to User
        self._api_keys: dict[str, MockMercutoIdentityService._GeneratedUserApiKey] = {}

    def get_my_permissions(self) -> VerifyMyPermissions:
        if self._client._auth_method is None:
            raise MercutoHTTPException('Not authenticated', 401)
        header: dict[str, str] = {}
        self._client._auth_method.update_header(header)
        if (api_key := header.get('X-Api-Key')) is not None:
            known = self._api_keys.get(api_key)
            if known is None:
                raise MercutoHTTPException('Invalid API key', 403)
            if known.hidden.custom_policy is not None:
                return VerifyMyPermissions(user=known.user, acl_policy=known.hidden.custom_policy)
            user = self._users[known.user]
            group = self._permission_groups[user.permission_group]
            return VerifyMyPermissions(user=user.code, acl_policy=group.acl_policy)
        elif (service_token := header.get('X-Service-Token')) is not None:
            if self._verify_service_token is None:
                raise MercutoHTTPException('Service token verification not implemented', 501)
            return self._verify_service_token(service_token)

        raise MercutoHTTPException('Invalid authentication method for mock implementation', 403)

    def list_users(self, tenant: Optional[str] = None) -> list[User]:
        if tenant is None:
            return list(self._users.values())
        return [u for u in self._users.values() if u.tenant == tenant]

    def create_user(
        self,
        username: str,
        tenant: str,
        description: str,
        group: str,
        default_password: Optional[str] = None
    ) -> User:
        code = str(uuid.uuid4())
        user = User(code=code, username=username, description=description, tenant=tenant, permission_group=group)
        self._users[code] = user
        self._user_details[code] = UserDetails(code=code, username=username)
        return user

    def get_current_user(self) -> CurrentUser:
        perms = self.get_my_permissions()
        if perms.user is None:
            raise MercutoHTTPException('Not authenticated', 401)
        user = self._users[perms.user]
        tenant = self._tenants[user.tenant]
        group = self._permission_groups[user.permission_group]
        return CurrentUser(
            code=user.code,
            username=user.username,
            description=user.description,
            tenant=tenant,
            permission_group=group,
            current_permission_policy=group.acl_policy
        )

    def get_user(self, code: str) -> User:
        return self._users[code]

    def delete_user(self, code: str) -> None:
        if code in self._users:
            del self._users[code]
        if code in self._user_details:
            del self._user_details[code]
        for k, v in list(self._api_keys.items()):
            if v.user == code:
                del self._api_keys[k]

    def get_user_details(self, code: str) -> UserDetails:
        return self._user_details[code]

    def set_user_details(
        self,
        code: str,
        email_address: Optional[str] = None,
        mobile_number: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None
    ) -> UserDetails:
        user_details = self._user_details[code]
        if email_address is not None:
            user_details.email_address = email_address
        if mobile_number is not None:
            user_details.mobile_number = mobile_number
        if first_name is not None:
            user_details.first_name = first_name
        if last_name is not None:
            user_details.last_name = last_name
        return user_details

    def get_user_api_keys(self, user: str) -> list[HiddenUserAPIKey]:
        return [v.hidden for v in self._api_keys.values() if v.user == user]

    def generate_api_key_for_user(
        self,
        user: str,
        description: str,
        custom_policy: Optional[str] = None
    ) -> VisibleUserAPIKey:
        key_code = str(uuid.uuid4())
        new_api_key = str(uuid.uuid4())
        hidden_key = HiddenUserAPIKey(code=key_code, description=description, last_used=None, custom_policy=custom_policy)

        if user not in self._users:
            raise MercutoHTTPException('User not found', 404)

        self._api_keys[new_api_key] = self._GeneratedUserApiKey(
            user=user, api_key=new_api_key, hidden=hidden_key
        )
        return VisibleUserAPIKey(code=key_code, new_api_key=new_api_key, description=description, custom_policy=custom_policy)

    def list_tenants(self) -> list[Tenant]:
        return list(self._tenants.values())

    def get_tenant(self, code: str) -> Tenant:
        return self._tenants[code]

    def create_tenant(
        self,
        name: str,
        description: str,
        logo_url: Optional[str] = None
    ) -> Tenant:
        code = str(uuid.uuid4())
        tenant = Tenant(code=code, name=name, description=description, logo_url=logo_url)
        self._tenants[code] = tenant
        return tenant

    def get_permission_groups(self, tenant: Optional[str] = None) -> list[PermissionGroup]:
        if tenant is None:
            return list(self._permission_groups.values())
        return [g for g in self._permission_groups.values() if g.tenant == tenant]

    def create_permission_group(
        self,
        tenant: str,
        label: str,
        acl_policy: str
    ) -> PermissionGroup:
        code = str(uuid.uuid4())
        group = PermissionGroup(tenant=tenant, code=code, label=label, acl_policy=acl_policy)
        self._permission_groups[code] = group
        return group

    def get_permission_group(self, group: str) -> PermissionGroup:
        return self._permission_groups[group]
